%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS comprule.ql,v 1.12 1993/12/10 02:23:12 marku Exp


%% Support for 'Compiled Rules'.
%% 
?- op(830, xfy, (=>)).
?- op(810, xfy, (and)).

:- declare(begin_compiled_rule/3).
:- declare(end_compiled_rule/1).
:- declare(compiled_rule/2).
:- declare(compiled_rule/3).
:- declare(compiled_rule/4).

%% begin_compiled_rule(+Call, +Patt, +OptionList)
%%
%%   Declares a new compiled rule group that will be used
%%   to generate theorems of the form Patt.
%%
%%   The outermost functor of Call is the name of the group.
%%   (It is necessary for the name of the group to be an atom so that
%%    we can generate a unique functor name for the asserted clauses.)
%%
%%   Any members of OptionList must be of the following form:
%%      precondition=Code

begin_compiled_rule(Call, Form, Options) :-
	Call =.. [Name|Args],
	available(Name),
	check_term(Form),
	!,

	%% check that Form does not contain quantifiers or object
	%% variables and that all its meta-variables appear in Call.
	check_or_writelist(
	    '$collect_variables'(Form, object_var, [], []),
	    ['Error: quantifiers/object-vars not allowed in: ',Form,nl]
	),
	check_or_writelist(
	    ('$collect_variables'(Form, meta_var, [], Vars),
	     '$collect_variables'(Call, meta_var, [], CallVars),
	     subset_var(Vars, CallVars)
	    ),
	    ['Error: all variables in: ', Form,nl,
	     '     must also occur in: ', Call,nl]
	),
	!,

	%% construct the internal call, by prefixing 'cr__' to name
	%% and adding one last argument for the theory name.
	atom_chars(Name, NameStr),
	current_dir(Theory),
	atom_chars(Theory, TheoryStr),
	atom_chars('cr_', [C,R,U]),
	append([C,R,U,U], TheoryStr, S1),
	append(S1, [U], S2),
	append(S2, NameStr, S3),
	atom_chars(IName, S3),
	append(Args,[_],Args2),
	InternalCall =.. [IName|Args2],
	!,

 	%% separate out the precondition
 	(
 	    delete(precondition = PreCond, Options, Options2)
 	;
 	    PreCond = true,
 	    Options2 = Options
 	),
	check_or_writelist(Options2 = [],
	    ['Unknown options: ', Options2]
	),

	%% assert this compiled rule into the current theory.
	functor(InternalCall, IName, IArity),
	assert_thy(save_clauses, [IName, IArity, IArity]),
	assert_thy(compiled_rule, [Name, Call, Form, _, open,
				   PreCond, InternalCall, _]).



%% end_compiled_rule(+Name).
%%
%%   This command completes the construction of the compiled rule 
%%   group called Name.  Once this command has been executed, no
%%   more rules can be added to the group.
%%
%%   Eventually, this command may cause the compiled rule group
%%   to *really* be compiled, rather than just asserted.

end_compiled_rule(Name) :-
	Args = [Name, Call, Form, Options, open, Pre, ICall, Vars],
	get_up_dir(compiled_rule, Args, Thy),
	check_or_writelist(
	    current_dir(Thy),
	    ['Error: begin/end_compiled_rule must be in the same theory',nl]
	),
	retract_thy(Thy, compiled_rule, Args),
	Args1 = [Name, Call, Form, Options, closed, Pre, ICall, Vars],
	assert_thy(Thy, compiled_rule, Args1).



%% get_compiled_rule(?Thy, ?Name, ?Call, ?Form)
%%
%%   This procedure gets the most commonly needed information about
%%   a compiled rule from the database, without evaluating the rule.
%%
get_compiled_rule(Thy, Name, Call, Form) :-
	get_up_dir(compiled_rule, [Name, Call, Form, _, _, _, _, _], Thy).



%% eval_compiled_rule(?Thy, ?Name, +Call, ?InForm, -FinalForm, -Deps).
%% eval_compiled_rule(?Thy, ?Name, ?Call, +InForm, -FinalForm, -Deps).
%% 
%%   This procedure attempts to generate a theorem
%%   that matches the specified arguments of Call and/or Form.
%%   If it succeeds, it instantiates Call and Form as well as
%%   returning the whole theorem in FinalForm.
%%   Deps is a list of dependency variables, to indicate which
%%   conjectures this compiled rule depends upon.

eval_compiled_rule(Thy, Name, Call, InForm, FinalForm, Deps) :-
	get_up_dir(compiled_rule, [Name,Call,Form,_,Open,Pre,ICall,_], Thy),

	%% Deduce Form from FinalForm, if Form is not specified.
	(var(InForm),nonvar(FinalForm) ->
	    %% this should return the conditional direction if possible.
	    choose_conditional(_, FinalForm, Conds, Form),
	    % precalculate the expected list of conditions (if nonvar).
	    (nonvar(Conds), Conds \== true ->
		gen_conjuncts(CList0, Conds)
	    ;
	        true
	    )
	;
	    Form = InForm
	),
	!,
	call(Pre),
	ipSet(cr_dependencies, []),
	ipSet(cr_conditions, CList0),
	current_dir(CD),
	!,
	call(ICall),
	ipLookup(cr_dependencies, Deps),
	ipLookup(cr_conditions, CList1),
	once(append(CList1, [], CList2)),

	% this may be either generating Conds from CList2, or
	% just checking CList2 against an instantiated Conds.
	gen_conjuncts(CList2, Conds),
	choose_conditional(_, FinalForm, Conds, Form).


%% compiled_rule(+Thm, +Name)
%% compiled_rule(+Thm, +(Call :- Code))
%% compiled_rule(+Thm, ?Dir, +Name)
%% compiled_rule(+Thm, ?Dir, +(Call :- Code))
%% compiled_rule(+Thm, ?Dir, +Name, ?(Call :- Code))
%%
%%   Compiles the theorem/axiom called Thm into a Qu-Prolog
%%   clause that becomes part of the definition of the compiled rule
%%   group called Name.  The resulting clause is asserted after
%%   the clauses generated by all previous compiled_rule commands.
%%
%%   Thm must be an axiom or a theorem, not a conjecture/hypothesis
%%   or another compiled rule (because the generation of that 
%%   compiled rule may use conjectures/hypotheses if this command
%%   is called from within a proof).
%%
%%   The second form allows the compiled clause to be specified more
%%   precisely.  Call must match the call term specified in the
%%   declaration of a compiled rule group.  The parameters within
%%   Call may be instantiated.  Code must be a comma-separated sequence
%%   of terms that specify how the conditions and body of Thm are
%%   to be discharged.  The allowable forms for these terms include:
%%   (where N is the number of some condition of Thm, starting from 1)
%%      body       -- 
%%      N=is_hyp   -- discharge condition N using is_hyp(_, _).
%%      N=return   -- discharge condition N by returning it as a
%%                      condition of the resulting theorem.
%%      N=Call2    -- discharge condition N by a recursive call
%%                    to some compiled rule group (not nec. this one).
%%      N          -- an abbreviation of N=Call2, where Call2 is a
%%                      call to *this* compiled rule group.
%%      N=Thm      -- discharge condition N at compile time, using the
%%                    rule called Thm.  Thm must be a fully qualified
%%                    name of the form Thy:Name and the rule that it refers
%%                    must not have any conditions, but may have a body.
%%      N=is_theorem -- discharge condition N by searching (at eval time)
%%                    for a theorem that matches condition N.
%%      {Term}     -- check that Term succeeds.
%%                      (generates code of the form:  \+ \+ Term)
%%
%%      various safe Qu-Prolog/Ergo predicates, such as
%%          var(_), nonvar(_), !, incomplete_retry_delays...
%%          (see gen_code/3 for details of what is allowed).

compiled_rule(Thm, OtherArgument) :-
	compiled_rule(Thm, _, OtherArgument).

compiled_rule(Thm, Dir, Name) :-
	atom(Name), !,
	compiled_rule(Thm, _, Name, _).

compiled_rule(Thm, Dir, (Call :- Code)) :-
	functor(Call, Name, _), !,
	compiled_rule(Thm, Dir, Name, (Call :- Code)).

compiled_rule(Thm, Name, (Call :- Code)) :-
	functor(Call, Name, _), !,
        compiled_rule(Thm, _Dir, Name, (Call :- Code)).


compiled_rule(Thm, Dir, Name, (Call :- Code)) :-
	get_up_dir(compiled_rule,
		   [Name,Call,Form,_,Open,_,ICall,_],
		   Theory),
	check_or_writelist(
	    Open == open,
	    ['Error: compiled_rule ', Name, ' is closed.',nl]
	),
	cr_gen_clause(Thm, Dir, Theory:Name, Form, 
		(Call :- Code),
		(ICall :- FinalCode)),

	%% ensure that this theory has a 'save_clauses' fact for ICall.
	current_dir(CD),
	functor(ICall, IName, IArity),
	(([save_clauses,CD]:[IName,IArity,_]) ->
	    update_dir(CD)
	;
	    assert_thy(CD, save_clauses, [IName, IArity, IArity])
	),

	\+ \+ (thaw((ICall, FinalCode)),
        writelist([ICall,' :-',nl,tab,FinalCode,'.',nl]),
	assertz((ICall :- FinalCode))).


%% cr_gen_clause(+ThmName, ?Dir, +CRName, +Form, +InClause, ?OutClause).
%%
%%   CRName must be the full name (ie. Thy:Name) of the compiled rule.
%%   Form must be the skeleton (pattern) of that compiled rule
%%   (and must have been instantiated to correspond to Call).
%%   InClause is the given (Call :- BodySpec).
%%   OutClause is the generated (InternalCall :- BodyCode).
/*******************
	((ThmForm = (Conds => Concl) ->
	    make_cond_list(Conds, 1, CList))
	;
	    (CList = [],
	     Concl = ThmForm)
	),
	Concl = Form,
*******************/


cr_gen_clause(Thm, Dir, Theory:Name, Form, (_Call :- Code0), (ICall :- Code)) :-

	%% determine what conditions must be discharged.
	find_rule(Thm, ThmForm, Body),

        once( choose_conditional(CList, ThmForm, Conds, ThmConcl)),
        once((best_direction(Dir, Form, ThmConcl),
	      choose_direction(Dir, ThmConcl, F, TEq, G),
	      choose_direction(FDir,   Form,  F, Eq, G),
	      member(FDir, [f, t]),
	      stronger(TEq, Eq),
	      incomplete_retry_delays)),

	make_cond_list(Conds, 1, [], CList),
	check_term(Form),
	!,


	((Body == true ->
	    CList2 = CList)
	;
	    (CList2 = [body=Body | CList])
	),

	%% generate Code from Code0
	gen_code_seq(Code0, CList2, info(Name), Code1),
	incomplete_retry_delays,
	!,

	%% mark this clause as belonging to the current theory.
	current_dir(CD),
	functor(ICall, _IName, IArity),
	arg(IArity, ICall, CD),

	%% add code to disable the clause if the compiled rule
	%% is called from a theory above CD.
	(CD == Theory ->
	    Code = Code1
	;
	    Code = (ancestor(CD),Code1)
	).


%% best_direction(?Dir, +Form, +ThmForm)
%%
%% Tweak the default direction depending on the theorem form and
%% the form of the compiled rule.
%%
best_direction(Dir, _F, _TF) :-
	nonvar(Dir), !.

best_direction(t, Form, ThmForm) :-
	functor(Form, F, _),
	\+ preorder(F).

best_direction(_Dir, _F, _TF).
	

%% make_cond_list(+Conjuncts, +Num, +List, ?CondList)
%%
%%   This turns the antecedents of the rule into a numbered list.
%%   eg.  make_cond_list(A and B, 1, [2], [[2,1]=A, [2,2]=B]).

make_cond_list(true,  1, _, []).

make_cond_list(Conds, N, NL, [RL=C|CList]) :-
	nonvar(Conds),
	Conds = (C and CS),
	!,
	N1 is N + 1,
	rev_tr([N|NL], RL),
	make_cond_list(CS, N1, NL, CList).

make_cond_list(C, N, NL, [RL=C]) :-
	rev_tr([N|NL], RL).
	%% C is not a conjunction.


%% gen_conjuncts(+CondList, -Conjunction).
%% gen_conjuncts(-CondList, +Conjunction).
%%

gen_conjuncts(L, C) :-
	L == [],
	!,
	C = true.

gen_conjuncts([C], C) :-
	C \== (_ and _),
	!.

gen_conjuncts([C|CL], C and CS) :-
	gen_conjuncts(CL, CS).



%% gen_code_seq(+CodeSpec, +CondList, +Info, -Code).
%%
%%   This turns the code sequence specified by the user (CodeSpec)
%%   into executable Qu-Prolog code.  It ensures that the resulting
%%   code will fulfill the following soundness requirements whenever
%%   it executes without failing:
%%
%%      SR1: the generated code will discharge all conditions in CondList.
%%           (however, it is allowed to discharge them by adding them
%%           to the implicit parameter list 'cr_conditions' at runtime.
%%           They will be returned as conditions of the resulting theorem).
%%
%%      SR2: the generated code must collect the dependency information
%%           of any hypotheses/conjectures that it uses, into the
%%           implicit parameter list 'cr_dependencies'.
%%
%%      SR3: whenever the generated code instantiates a variable that 
%%           may appear in the final theorem, the instantiated term 
%%           must be legal within the theory that this compiled_rule/2
%%           command was executed from.
%%
%%      SR4: the generated code will not have any side effects on the 
%%           current proof, or on any implicit parameters (except for
%%           cr_conditions and cr_dependencies, as mentioned above).
%%           This implies that it cannot call arbitrary
%%           Prolog predicates (eg. trans, h_trans, conjecture).
%%           (NB. it does not check for side effects on the Prolog
%%            database, so we must assume that CodeSpec will not
%%            assert/retract clauses.  A Prolog module system
%%            is needed to fix this properly).

gen_code_seq(Spec, CList, Info, Code) :-
	var(Spec),
	!,
	gen_all_discharge_code(CList, Info, Code).

gen_code_seq((Spec,Rest), CList, Info, (Code,CodeRest)) :-
	!,
	gen_code_frag(Spec, CList, Info, CList2, Code),
	!,
	gen_code_seq(Rest, CList2, Info, CodeRest).

gen_code_seq(Spec, CList, Info, Code) :-
	%% nonvar(Spec), Spec \= (_,_)
	!,
	gen_code_frag(Spec, CList, Info, CList2, Code),
	check_or_writelist(
	    CList2 = [],
	    ['Error: the following conditions were not discharged: ',nl,
	      tab,CList2,nl]
	).


%% gen_all_discharge_code(+CondList, +Info, -Code)
%%
%%   Generates Code to discharge all the remaining conditions in CondList.

gen_all_discharge_code([], _, true) :-
	!.

gen_all_discharge_code([N = C], Info, Code) :-
	!,
	gen_discharge_code(N, C, Info, _, Code, []).

gen_all_discharge_code([N = C | CList], Info, (Code,Rest)) :-
	gen_discharge_code(N, C, Info, _, Code, []),
	gen_all_discharge_code(CList, Info, Rest).
 

%% gen_code_frag(+Spec, +CList0, +Info, -CList, -Code).
%%
%%   Generates Code from Spec.  If Spec specifies code to
%%   discharge a condition, that condition is removed from
%%   CList0 to give CList, otherwise CList0 and CList are
%%   identical.

gen_code_frag(N, CList0, Info, CList, Code) :-
	(N = (Num0=Spec) ; Num0 = N),
	(integer(Num0) 
	 ->
		Num = [Num0]
	 ;
		(isList(Num0) ; Num0 == body),
		Num = Num0
	),
	!,
	check_or_writelist(
	    delete(Num = Cond, CList0, CList1),
	    ['Error: ', Num=Cond, ' discharged too many times.',nl]
	),
	gen_discharge_code(Num, Cond, Info, Spec, Code, NewConds),
	append(NewConds, CList1, CList).

gen_code_frag(Spec, CList0, Info, CList0, Code) :-
	gen_code(Spec, Info, Code).



%% gen_discharge_code(+N, +Cond, +Info, ?Spec, -Code, ?NewConds).
%%
%%   Generates Code to discharge Cond, where Cond is the N'th
%%   condition of the rule being compiled (or if N == body,
%%   then Cond is the Qu-Prolog side condition of that rule).
%%   Spec determines how the condition should be discharged.
%%
gen_discharge_code(body, Code, _Info, _, Code, []) :-
	!.

gen_discharge_code(_N, Cond, info(Name), Spec, ICall, []) :-
	var(Spec),
	!,
	%% recursive call to this compiled rule.
	get_up(compiled_rule, [Name, _, Form, _, _, _, ICall, _]),
	Cond = Form,
	incomplete_retry_delays.


%% Are the following two clauses in the desired order?
%%
gen_discharge_code(_N, Cond, _Info, is_hyp, cr_is_hyp(Cond), []) :-  
	!.

gen_discharge_code(_N, Cond, _Info, return, cr_return(Cond), []) :-
	!.

gen_discharge_code(_N, Cond, _Info, is_theorem, is_theorem(_,Cond), []) :-
	!.

gen_discharge_code(N, Cond, _Info, Thy:Name, Body, NewConds) :-
	!,
	check_or_writelist(
	    find_rule_at2(Name, ThmForm, Body, Thy),
	    ['Error: Rule ', Thy:Name, ' not found.', nl]),
	check_or_writelist(
	    (choose_conditional(NewConds, ThmForm, ThmCond, ThmConcl),
	     discharges(ThmConcl, Cond)),
	    ['Error: Rule ', Thy:Name, ' does not discharge condition ', N=Cond, nl]),
	make_cond_list(ThmCond, 1, N, NewConds), !.
	

gen_discharge_code(_N, Cond, _Info, Call, ICall, []) :-
	functor(Call, Name, _),
	get_up(compiled_rule, [Name, Call, Form, _, _, _, ICall, _]),
	Cond = Form,
	incomplete_retry_delays,
	!.

gen_discharge_code(N, Cond, _Info, Spec, _, _) :-
	writelist(['Error: condition number ', N, ': ', Cond,nl,
		'     cannot be discharged by: ', Spec,nl]),
	fail.


%% discharges(+T, +T1)
%%
%%	Unfortunate workaround; T => T1
%%
discharges(T, T1) :-
	choose_direction(_, T, F, E, G),
	choose_direction(_, T1, F, E1, G),
	stronger(E, E1), !.

%% cr_is_hyp(+Cond)
%%
%%   Auxillary function for discharging a condition, Cond, by
%%   finding a matching hypothesis (and recording its dependencies).
%%
cr_is_hyp(Form) :-
	ipLookup(hyplist, HL),
	member((_, Form, Del, _, Dep), HL),
	var(Del),
	incomplete_retry_delays,
	(Dep == [] ->
	    true
	;
	    ipLookup(cr_dependencies, DepList0),
	    '$collect_variables'(Dep, var, DepList0, DepList),
	    ipSet(cr_dependencies, DepList)
	).


%% cr_return(+Cond)
%%
%%   Auxillary function for discharging a condition, Cond, by
%%   returning it as a condition of the resulting theorem.
%%
%%   TODO: this should perhaps use member_var to match against
%%         existing members of cr_conditions.  However, it must
%%         still be able to instantiate/extend cr_conditions in
%%         order to add new conditions.
%%         
cr_return(Cond) :-
	ipLookup(cr_conditions, CList),
	once(member(Cond,CList)).



%% gen_code(+Spec, +Info, -Code).
%%
%%   Generates Code from Spec.  This procedure handles all Spec's 
%%   that are not discharging conditions.
%%
%%   All code generated by this procedure must satisfy the 
%%   soundness constraints, SR3 and SR4, mentioned above.

gen_code({T},                     _, (\+ \+ T)) :- !.
gen_code(\+ T,                    _, \+ T) :- !.

gen_code(!,                       _, !) :- !.
gen_code(A==B,                    _, A==B) :- !.
gen_code(A\==B,                   _, A\==B) :- !.
gen_code(true,                    _, true) :- !.
gen_code(fail,                    _, fail) :- !.
gen_code(var(A),                  _, var(A)) :- !.
gen_code(nonvar(A),               _, nonvar(A)) :- !.
gen_code(is_object_var(A),        _, is_object_var(A)) :- !.
gen_code(integer(A),              _, integer(A)) :- \+ functor(A,_,_), !.
gen_code(cr_freeze(A),            _, true) :- freeze(A), !.
gen_code(atom(A),                 _, atom(A)) :- !.
gen_code(atomic(A),               _, atomic(A)) :- !.
gen_code(isList(A),               _, isList(A)) :- !.
gen_code(frozen(A),               _, frozen(A)) :- !.
gen_code(msg(M),                  _, msg(M)) :- !.    %% ????????????
gen_code(less_verbose(msg(M)),    _, less_verbose(msg(M))) :- !.
gen_code(less_verbose(N,msg(M)),  _, less_verbose(N,msg(M))) :- !.
gen_code(incomplete_retry_delays, _, incomplete_retry_delays) :- !.

gen_code(Spec, _, _) :-
	writelist(['Error: unknown code specification: ',Spec,nl]),
	fail.
