%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS iu.ql,v 1.8 1993/12/10 02:23:34 marku Exp


?- op(600,xfy,nfi).  %% same as not_free_in
?- object_var(x).

/*
 * with_frozen_nfi_test(+Flag, +Code)
 *
 *  Executes Code with frozen_nfi_test flag set to Flag.
 *  It fails iff Code fails.
 *  It always resets the frozen_nfi_test to its original value.
 */

with_frozen_nfi_test(Flag, Code) :-
    setFlag(frozen_nfi_test, Flag, Orig),
    Code,
    setFlag(frozen_nfi_test, Orig, _).


/*
 * A user-level predicate for adding not_free_in conditions.
 */

nfi(Var,Term) :-
    ipLookup(nfi, L),
    ipSet(nfi, [(Var, Term) | L]),
    with_frozen_nfi_test(off, Var not_free_in Term).



/* complete_retry_delays.
 *
 *  Retries all delayed problems and fails if
 *  they cannot all be solved.
 */

complete_retry_delays :-
    '$check_delayed_problem'('='), !,
    retry_delay_problems,
    \+ '$check_delayed_problem'('=').

complete_retry_delays.


/*
 * incomplete_retry_delays
 *
 * attempts to solve all the delayed unification problems.
 * fails if it can't find any solution.
 * 
 * incomplete_retry_delays makes extensive use of incomplete_unify
 *
 * NOTE that throughout use is made of the frozen_nfi_test flag.
 * If this flag is set then unify fails if any not_free_in problems
 * involving frozen object-variables are generated.
 * This forces incomplete_unify to ignore possible unifiers that
 * generate such not_free_in problems.
 */

incomplete_retry_delays :-
	retry_delay_problems,
	get_unify_delays(Delays),
	(   Delays == []
	    ->
		true
	    ;
	        '$getFlag'(frozen_nfi_test,Flag),
		writeln('incomplete_retry_delays with frozen_nfi_test' = Flag),
		write_delays_list(Delays),
		get_best_delay(Delays, T1, T2),
		incomplete_unify(T1, T2),
		incomplete_retry_delays
	).

write_delays_list([]) :-
	nl.
write_delays_list([D|DS]) :-
	tab(8),writep(D),nl,
	write_delays_list(DS).


/*
 * incomplete_unify
 *
 * The aim of incomplete_unify is to solve a delayed
 * unification problem by searching through possible
 * ways of producing unifiers for this delayed problem.
 *
 *
 * The possible delayed unification problems fall into
 * the following groups (up to symmetry).
 * (In this list A,B,... are meta-variables x,y, ...
 *  are object-variables and s, s1, s2, ...  are substitutions)
 *
 * (a) s*A = structured_term  (including constant)
 *             (s /= [])
 *
 * (b) s*x = structured_term  (including constant)
 *             (s /= [])
 *
 * (c) s1*A = s2*B    (not (s1 = [] & s2 = [])
 *
 * (d) s1*x = s2*y    (not (s1 = [] & s2 = [])
 *
 * (e) s1*x = s2*A    (not (s1 = [] & s2 = [])
 *
 * 
 * The method used to find unifiers is to take a delayed
 * problem, make a simplifying choice (as given below)
 * and then rely on retry_delay (and hence unify) to do
 * the real work of finishing off the problem.
 * In some cases, a simplifying choice will not cause the
 * current problem to be solved. In this case backtracking
 * occurs and a more extensive simplifying choice is made.
 *
 * Simplifying choices for delayed problems
 * ----------------------------------------
 *
 * These follow the problem types above. NOTE that reordering
 * these choices will cause different solutions to appear
 * in different orders.
 * The choices given should cover all the possible ways
 * in which a solution can be found.
 *
 * In most of these cases failure might occur, the problem could
 * be completly solved, or new (simpler) unification problems
 * may be produced.
 *
 * (a) 1. A = t  where t is a 'skeletal' form of the structured term
 *               (e.g. if the structured term is f(t1, t2) then
 *                     t is f(A1, A2) for new meta-variables)
 *     2. A = x  where x is a new object variable.
 *
 * (b) Assume s = [t1/x1,....,tn/xn]
 *
 *     x not_free_in xj (j = i+1,...,n) AND x = xi (leave out x = xi if i=0)
 *        (i = 1..n in turn)
 *
 * (c) Assume s1 = [t1/x1,....,tn/xn], s2 = [u1/y1,....,uk/yk]
 *
 *     1. if A /== B and A is not frozen then delete some
 *        entries from s1 and s2 then bind A to s2' * B
 *        (s2' is s2 with entries removed)
 *
 *     2. if A /== B and B is not frozen do the flip of 1.
 *
 *     3. if A == B  force s1 and s2 to have the same effect on A.
 *
 *     4. A = x (new x)
 *     
 *     5. B = x (new x)
 *
 * (d) Assume s1 = [t1/x1,....,tn/xn], s2 = [u1/y1,....,uk/yk]
 *
 *     1. x not_free_in xj (j = i+1,...,n) AND x = xi (leave out x = xi if i=0)
 *        (i = 1..n in turn)
 *   
 *     2. y not_free_in yj (j = i+1,...,k) AND y = yi (leave out y = yi if i=0)
 *        (i = 1..k in turn)
 *
 *     3. y = x   +   force s1 and s2 to have the same effect on x.
 *
 * (e) Assume s1 = [t1/x1,....,tn/xn]
 *
 *     1. x not_free_in xj (j = i+1,...,n) AND x = xi (leave out x = xi if i=0)
 *        (i = 1..n in turn)
 *
 *     2. delete some entries from s1 and s2 (until s2 is empty) then bind
 *        A to s1'*x
 *
 *     3. A = z (for some new object variable)
 *        THEN do d(1,3)  - i.e. in d1 z is the variable used where x appears.
 *        NOTE that d2 is covered in case e1 followed by e2.
 *   
 *
 *
 * In some of the above cases ((c) and d3) if the particular modification
 * of the substitutions does not produce a solution for this problem
 * then failure should occur and a different modification should be tried.
 * If this is not done then an infinite loop will occur.
 * This problem can be avoided by using first_delay_problem/1 to compare
 * the current problem with the top delay problem (after retry_delays)
 * If the current problem is the same as the top delay problem
 * then no progress is made and so a different modification should be tried.
 *
 * In c1, c2, c4 and d3 there are several possibilities for deleting
 * and matching components of substitutions.
 *
 * Assume the problems are [t1/x1,....,tn/xn]*A and [u1/y1,....,uk/yk]*B
 * (same for A being x and B being y)
 *
 * 1. xi not_free_in A
 * 2. yi not_free_in B
 * 3. ti = xi
 * 4. ui = yi
 * 5. if xi == yj , ti = uj
 * 6. if xi /== yj , xi = yj  AND ti = uj
 *
 * Any single attempt to solve c1,c2,c4 or d3 will involve making
 * several of these choices.
 */


incomplete_unify(T1, T2) :-
	simplify_term(T1, ST1), /* sequence of subs -> single sub */
	simplify_term(T2, ST2),
	setFlag(frozen_nfi_test, on, Old),
	iu(ST1, ST2),
	setFlag(frozen_nfi_test, Old, _).

iu(T1, T2) :-
	var(T1),
	structure(T2),
	!,
	iu_case_a(T1, T2).
iu(T1, T2) :-
	var(T2),
	structure(T1),
	!,
	iu_case_a(T2, T1).
iu(T1, T2) :-
	is_object_var(T1),
	structure(T2),
	!,
	iu_case_b(T1, T2).
iu(T1, T2) :-
	is_object_var(T2),
	structure(T1),
	!,
	iu_case_b(T2, T1).
iu(T1, T2) :-
	var(T1),
	var(T2),
	!,
	iu_case_c(T1, T2).
iu(T1, T2) :-
	is_object_var(T1),
	is_object_var(T2),
	!,
	iu_case_d(T1, T2).
iu(T1, T2) :-
	is_object_var(T1),
	var(T2),
	!,
	iu_case_e(T1, T2).
iu(T1, T2) :-
	is_object_var(T2),
	var(T1),
	!,
	iu_case_e(T2, T1).

/********  Case a *********/

iu_case_a(X, T) :-
	break_term(X, SX, BX),
	((   
		\+ var_occurs_in(BX, T),
	        skeleton(T, BX)      			/* case a1 */
	    ;
		BX = x             			/* case a2 */
	)).

/********  Case b *********/

iu_case_b(X, T) :-
	break_term(X, SX, BX),
	force_deref(SX, BX).

/********  Case c *********/

iu_case_c(X, Y) :-
	break_term(X, SX, BX),
	break_term(Y, SY, BY),
	((
		BX \== BY,
		\+ (frozen(BX)),			/* case c1 */
		((frozen(BY) ; SY \== [])),
		strip_sub(SX, BX, SY, NSY),
		build_term(NY, NSY, BY),
		BX = NY,
		successful_bind(BX,NY),
		make_progress(X,Y)
	    ;
		BX \== BY,
		\+ (frozen(BY)),			/* case c2 */
		((frozen(BX) ; SX \== [])),
		strip_sub(SY, BY, SX, NSX),
		build_term(NX, NSX, BX),
		BY = NX,
		successful_bind(BY,NX),
		make_progress(X,Y)
	    ;
		BX == BY,				/* case c3 */
		pair_subs(BX, SX, SY, NSX, NSY),
		make_progress(X,Y)
	    ;
		\+ (BX == BY),
		BX = x,				/* case c4 */
		force_deref(SX, BX)  
	    ;
		\+ (BX == BY),
		BY = x,				/* case c5 */
		force_deref(SY, BY)  
	)).

/********  Case d *********/

iu_case_d(X, Y) :-
	break_term(X, SX, BX),
	((
	        force_deref(SX, BX)		/* case d1 */
	    ;
	        break_term(Y, SY, BY),
		((
	                force_deref(SY, BY),  	/* case d2 */
			reduce_range(Y, X, SX, BX) /* avoid repeated cases */
		    ;
			BX = BY,			/* case d3 */
			pair_subs(BX, SX, SY, _NSX, _NSY),
			make_progress(X,Y)
		))
	)).


/********  Case e *********/

iu_case_e(X, Y) :-
	break_term(X, SX, BX),
	((
		force_deref(SX, BX)		/* case e1 */
	    ;
		break_term(Y, SY, BY),
		((
			SX \== [],
			strip_sub(SY, BY, SX, NSX),	/* case e2 */
			build_term(NX, NSX, BX),
			BY = NX,
			successful_bind(BY,NX),
			make_progress(X,Y)
		    ;
			BY = x,			/* case e3 */
			((
				force_deref(SY, BY)
			    ;
				BX = BY,
				pair_subs(BX, SX, SY, _NSX, _NSY),
				make_progress(X,Y)
			))
		))
	)).
		
/*
 * If the binding of Var to VarWithSub delays
 * then it must be because Var occurs in the sub of VarWithSub.
 * sucessful_bind attempts to remove such occurrences from consideration
 * by adding appropriate not_free_in constraints
 * 
 */

successful_bind(Var,VarWithSub) :-
	make_progress(Var,VarWithSub),!.
successful_bind(Var,VarWithSub) :-
	break_term(VarWithSub, SV, V),
	fix_occurs_check(Var, SV, V).

fix_occurs_check(_,[],_).
fix_occurs_check(Var, [(R/D)|Sub], V) :-
	\+ \+ var_occurs_in(Var, R),
	!,
	D not_free_in V,
	fix_occurs_check(Var, Sub, V).
fix_occurs_check(Var, [(_,_)|Sub], V) :-
	fix_occurs_check(Var, Sub, V).


var_occurs_in(V, Term) :-
	'$collect_variables'(Term, var, [], VarList),
	is_in(V, VarList).

is_in(V, [H|T]) :-
	V == H,!.
is_in(V, [_|T]) :-
	is_in(V, T).
	
	

/*
 * reduce_range is used to partially reduce the subs in a range
 * entry to avoid repeating cases
 * reduce_range(?,?,?)
 */

reduce_range(R, _ ,_ , _) :-
	structure(R),  	/* case handled in case b */
	!,
	fail.
reduce_range(Y, X, SX, BX) :-
	var(Y),        /* avoid e1 case */
	!,
	break_term(Y, SY, BY),
	((
		strip_sub(SY, BY, SX, NSX),	/* case e2 */
		build_term(NX, NSX, BX),
		BY = NX,
		successful_bind(BY,NX),
		make_progress(X,Y)
	    ;
		SY = x,			/* case e3 */
		((
			force_deref(SY, BY)
		    ;
			BX = BY,
			pair_subs(BX, SX, SY, _NSX, _NSY),
			make_progress(X,Y)
		))
	)).
reduce_range(Y, X, SX, BX) :-
	is_object_var(Y),        /* avoid d1 case */
	!,
	break_term(Y, SY, BY),
	((
		force_deref(SY, BY),  	/* case d2 */
		reduce_range(Y, X, SX, BX)
	    ;
		BX = BY,			/* case d3 */
		pair_subs(BX, SX, SY, _NSX, _NSY),
		make_progress(X,Y)
	)).

/*
 * break up a sub*term into its components flattening out the sub
 * break_term(?,?,!)
 */

break_term(T, ST, BT) :-
	substitute(T, S1, BT),
	flatten_sub(S1, ST).

flatten_sub([], []) :- !.
flatten_sub([X], X).

/*
 * build a term from sub and subterm
 * build_term(!,?,?)
 */

build_term(T, ST, BT) :-
	flatten_sub(S1, ST),
	substitute(T, S1, BT).

/*
 * build a skeleton of a structured term.
 * skeleton(?,!)
 */

skeleton(T, SkelT) :-
	T =.. [Op|Args],
	!,
	length(Args, N),
	length(NArgs, N),
	SkelT =.. [Op|NArgs].
skeleton(T, SkelT) :-
	quantify(T, Q, _, _),
	quantify(SkelT, Q, x, _).


/*
 * try all possible ways in which an object var might deref through
 * a sub - return the dereferenced term
 * force_deref(?,?)
 */

force_deref([], X) :- !, fail.
force_deref(Sub, X) :-
	rev_tr(Sub, RSub),
	force_deref_rev(RSub, X).

force_deref_rev([], X).
force_deref_rev([(R/X)|_], X).
force_deref_rev([(_/D)|T], X) :-
	X not_free_in D,
	force_deref_rev(T, X).


/*
 * strip_sub(SX, BX, SY, NSY) uses a combination of matching
 * entries in SX with SY and adding not_free_in constraints to BX
 * to remove SX - NSY is the remainder of SY
 */

strip_sub([], BX, SY, SY).
strip_sub([(R/D)|SX], BX, SY, NSY) :-
	is_not_free_in(D,BX),	/* no need in this case */
	!,
	strip_sub(SX, BX, SY, NSY).
strip_sub([(R/D)|SX], BX, SY, NSY) :-
	((
		R = D
	    ;
		D not_free_in BX
	)),
	strip_sub(SX, BX, SY, NSY).
strip_sub([(R/D)|SX], BX, SY, NSY) :-
	((
		iu_compatible(R,D,SY, SY1)
	    ;
		iu_delete(R,D, SY, SY1)
	)),
	strip_sub(SX, BX, SY1, NSY).

/*
 * pair_subs takes two (sub, var) pairs  (var could also be an object variable)
 * and attempts to remove some substitution entries in a way that
 * tries to mantain  the relative effects of the subs on their corresponding
 * vars.
 *
 * This is achieved by making domains not free in vars,
 * making domains match their corresponding range,
 * paring up an entry in s1 with one in s2.
 * pair_subs(?,?,?,!,!)
 * 
 * NOTE pair_subs is only used when the variable parts are ==.
 *
 * DRList keeps track of the domain elements in the first sub
 * that are set to the corresponding range element so that
 * delete_entries does not try not_free_in constraints
 * to eliminate a corresponding entry in the second sub
 * (this possibility is covered in the second lats pair_subs2 call)
 */


pair_subs(_X, [], [], [], []) :- !.
pair_subs(X, SX, SY, SX2, SY2) :-
	match_sub_entries(SX, SY, SX1, SY1),
	pair_subs1(X, SX1, SY1, SX2, SY2).

pair_subs1(X, [], SY, SX1, SY1) :-
        !,
        pair_subs2(X, SY, [], SY1, SX1, []).
pair_subs1(X, SX, SY, SX1, SY1) :-
        pair_subs2(X, SX, SY, SX1, SY1, []).

pair_subs2(X, [(R/D)|T], SY, NSX, NSY, DRList) :-
	is_not_free_in(D,X),	/* no need in this case */
	!,
        pair_subs2(X, T, SY, NSX, NSY, DRList).
pair_subs2(X, [(R/D)|T], SY, NSX, NSY, DRList) :-
        iu_compatible(R,D,SY,CSY),
        pair_subs2(X, T, CSY, NSX, NSY, DRList).
pair_subs2(X, [(D/D)|T], SY, NSX, NSY, DRList) :-  /* R = D */
        pair_subs2(X, T, SY, NSX, NSY, [D|DRList]).
pair_subs2(X, [(R/D)|T], SY, NSX, NSY, DRList) :-
        iu_delete(R,D, SY, SY1),
        pair_subs2(X, T, SY1, NSX, NSY, DRList).
% pair_subs2(X, [(R/D)|TX], SY, [(R/D)|NSX], NSY, DRList) :-
%         pair_subs2(X, TX, SY, NSX, NSY, [D|DRList]).
pair_subs2(X, [(R/D)|T], SY, NSX, NSY, DRList) :-
        \+ frozen(D),
        D not_free_in X,
        pair_subs2(X, T, SY, NSX, NSY, DRList).
pair_subs2(X, [], SY, [], NSY, DRList) :-
        delete_entries(SY,X, NSY, DRList).


/*
 * NOTE - we will miss unifiers if we have pairs in different orders
 * and the order is important
 * e.g. [A/x,B/y]*C = [B/y,A/x]*C with x, y not known to be apart
 */

match_sub_entries([], SY, [], SY).
match_sub_entries([(RX/DX)|SX], SY, SX1, SY1) :-
	delete_match(RX, DX, SY, SY0),
	!,
	match_sub_entries(SX, SY0, SX1, SY1).
match_sub_entries([HX|SX], SY, [HX|SX1], SY1) :-
	match_sub_entries(SX, SY, SX1, SY1).

delete_match(RX, DX, [(RY/DY)|SY], SY) :-
	DX == DY,
	RX == RY,
	!.
delete_match(RX, DX, [HY|SY], [HY|SY1]) :-
	delete_match(RX, DX, SY, SY1).


iu_compatible(R,D, [(RY/DY)|T], T) :-
        D == DY,
        R = RY.
iu_compatible(R,D, [E|T], [E|NT]) :-
        iu_compatible(R,D,T, NT).

iu_delete(R,D, [(RY/DY)|T], T) :-
        \+ (D == DY),
        D = DY,
        R = RY.
iu_delete(R,D, [E|T], [E|NT]) :-
        iu_delete(R,D,T, NT).


delete_entries([],_ , [], _).
% delete_entries([H|T],Y, [H|NT], DRList) :-
%         delete_entries(T,Y, NT, DRList).
delete_entries([(R/D)|T],Y, NT, DRList) :-
	is_not_free_in(D,Y),	/* no need in this case */
	!,
        delete_entries(T,Y, NT, DRList).
delete_entries([(D/D)|T],Y, NT, DRList):-    /* D = R */
        delete_entries(T,Y, NT, DRList).
delete_entries([(R/D)|T],Y, NT, DRList):-
        \+ frozen(D),
	\+ in_list(D, DRList),
        D not_free_in Y,
        delete_entries(T,Y, NT, DRList).

	
make_progress(A,B) :-
        setFlag(frozen_nfi_test, off,Old),
        retry_delay_problems,
        (    get_unify_delays(P) 
	     ->
                make_progress2(A,B,P)
            ;
		true
	),
	setFlag(frozen_nfi_test, Old,_).

make_progress2(A,B,L) :-
        \+ (found_problem(A,B,L)).

found_problem(A,B,[H|T]) :-
        H == (A = B) ; H == (B = A).
found_problem(A,B,[_|T]) :-
        found_problem(A,B,T).

structure(T) :-
	\+ var(T),
	\+ is_object_var(T).

in_list(A, [H|T]) :-
	A == H,
	!.
in_list(A, [_|T]) :-
	in_list(A,T).

/*
 * Delays that involve structures typically have the least
 * number of possible unifiers therefore they should be
 * unified first.
 * So get_best_delay trys to find such a problem
 */

get_best_delay(P, T1, T2) :-
	get_best_delay1(P, T1, T2),
	!.
get_best_delay([(T1 = T2)|_], T1, T2).

get_best_delay1([(T1 = T2)|_], T1, T2) :-
	((structure(T1) ; structure(T2))),
	!.
get_best_delay1([_|T], T1, T2) :-
	get_best_delay(T, T1, T2).


/*
 * is_not_free_in(X,T) is true iff
 * the object variable X is known to be not free in the variable
 * or object variable (without sub) T
 */

is_not_free_in(x, T) :-
	simplify_term([1/x]*T, ST),
	substitute(ST, [], BT),
	BT == T.
