%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS exprabbr.ql,v 1.6 1993/12/10 02:18:40 marku Exp


/*
 * expr_de_abbreviate(Ein, Eout)
 *
 *	This predicate produces Eout from Ein, by expanding all 
 *	abbreviations in Ein.  The abbreviations are defined by
 *	declarations of the form:
 *
 *		abbreviation E === E2.
 *
 *	WARNING: these abbreviations are expanded repeatedly,
 *	so you have to be careful not to define a series of
 *	abbreviations that can cause a loop!
 *
 *	Only atoms, functions and quantifiers can be abbreviations,
 *	not substitutions, meta-variables or object variables.
 *      Furthermore, all variables are frozen while expr_de_abbreviate
 *      executes so that they are not accidentally instantiated.
 */


?- object_var('x').
?- op(810, xfy, (and)).


expr_de_abbreviate(In, Out) :-
	%% For efficiency, we only collect the variables once.
	'$collect_variables'(In, var, [], Vars),
	'$freeze2'(Vars),
	expr_de_abbrev(In, Out),
	'$thaw2'(Vars).


/***********************************************************************
 * Auxillary clauses which should not be called from outside this module.
 ***********************************************************************/

expr_de_abbrev(In, Out) :-
	(
	    (substitute(In, Subs, Term),
	     \+ (Subs = [])
	    -> subslists_de_abbrev(Subs, SubsOut),
	       expr_de_abbrev(Term, TermOut),
	       substitute(Out, SubsOut, TermOut)
	    )
	;
	    (var(In)
	     /* and not preceeded by a substitution */
	    -> Out = In
	    )
	;
	    (is_object_var(In)
	     /* and not preceeded by a substitution */
	    -> Out = In
	    )
	;
	    (get_up(abbreviation,[In,Tmp])
	    -> expr_de_abbrev(Tmp, Out)
	    )
	;
	    (In = (A and B), \+ functor(A, and, _)
	    -> expr_de_abbrev(A, Aout),
	       expr_de_abbrev(B, Bout),
	       ((functor(Aout, and, 2) ->
		   arg(1, Aout, A1),
		   arg(2, Aout, A2),
		   Out = (A1 and A2 and Bout))
	       ;
	           (Out = (Aout and Bout))
	       )
	    )
	;
	    (functor(In, Func, Arity)
	    -> functor(Out, Func, Arity),
	       expr_de_abbrev_args(In, Out, Arity)
	    )
	;
	    (quantify(In, Q, Var, Body)
	     /* and not preceeded by a substitution */
	    -> expr_de_abbrev(Body, BodyOut),
	       quantify(Out, Q, Var, BodyOut)
	    )
	).


/* This de_abbreviates the arguments of In. */
expr_de_abbrev_args(In, Out, N) :-
	(
	  (N > 0
	  -> arg(N, In, Arg),
	     arg(N, Out, ArgOut),
	     expr_de_abbrev(Arg, ArgOut),
	     N1 is N - 1,
	     expr_de_abbrev_args(In, Out, N1)
	  )
	;
	  true
	).


/* This goes through substitution lists, deabbreviating expressions */
subslists_de_abbrev([], []).

subslists_de_abbrev([S|Subs], [SOut|SubsOut]) :-
	subs_de_abbrev(S, SOut),
	subslists_de_abbrev(Subs, SubsOut).

/* This goes through substitutions, deabbreviating expressions */
subs_de_abbrev([], []).

subs_de_abbrev([T/Var|Subs], [TOut/Var|SubsOut]) :-
	expr_de_abbrev(T, TOut),
	subs_de_abbrev(Subs, SubsOut).
