%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS it.ql,v 1.10 1993/12/10 02:18:48 marku Exp


%% tactic(+CallExpr, +Documentation, +WrapperLabel, +SubProofs).
%%
%%  Declares and documents a tactic.  Also determines what wrapper
%%  information goes into the proof record when the tactic is called
%%  using wrap_tac/1.
%%
%%  First two clauses are for backwards compatibility only.

tactic(FA, Label, ListofLabels, bounded) :-
	writelist(['WARNING: tactic ', FA, ' should be updated to use\n',
		   '   the new style of tactic declaration sometime.\n']),
	!,
	tactic(FA, 'not documented', Label, ListofLabels).

tactic(FA, _, _, unbounded) :-
	writelist(['ERROR: tactic ', FA, ' must be updated.']),
	!,
	fail.

tactic(FA, Doc, Label, ListofLabels) :-
	isList(ListofLabels),
	functor(FA, F, _),
	make_wrap_list(ListofLabels, Scope, WList, CList),
	tactic_assert(tactic, [F, Doc, FA, Label, Scope, WList, CList]).
				%% What is CList used for?


%% make_wrap_list(+LabelList, ?Scope, -WrapperList, -CodeList)
%%
%%  This turns a list of labels into a list of wrappers.
%%  If a tail of LabelList is a variable, the resulting WrapperList
%%  (and CodeList) are unbounded.

make_wrap_list(A, _, Tail, Tail) :-
	var(A),
	!.
	
make_wrap_list([], _, [], []).

make_wrap_list([A|B], S, [call_tac(- S, A, _)|WB], [_|CB]) :-
	make_wrap_list(B, S, WB, CB).


/*
**  Use call_tac for invisible tactic calls
**
**  Use wrap_tac for recorded (though perhaps hidden) tactic calls
*/

wrap_tac(Call) :-
	functor(Call, F, Arity),
	get_up(tactic, [F, Doc, FCall, Label, Scope, WList, _]),
	add_wrappers(Call, WList, FCall),
	!,
	call_tac(wrap(+ Scope, (Call, Label), FCall)).

wrap_tac(Call) :- 
	call_tac(Call).

% no wrappers needed, because no subproofs.
add_wrappers(Call, [], Call).

% Call and FCall have same arity, so put wrappers around each element
%   in the last argument of Call (the list of subproofs).
add_wrappers(Call, WList, FCall) :-
	WList \== [],
	functor(Call, F, N),
	functor(FCall, F, N),
	N1 is N - 1,
	same_args(N1, Call, FCall),
	arg(N, Call, Arg),
	fix_wrapped_code(Arg, WList),
	arg(N, FCall, WList).

% Call has one less argument than FCall, so add WList to the
%   arguments of Call (as the last argument), giving FCall.
add_wrappers(Call, WList, FCall) :-
	WList \== [],
	functor(Call, F, N1),
	functor(FCall, F, N),
	N1 is N - 1,
	same_args(N1, Call, FCall),
	arg(N, FCall, WList).


% same_args(+Num, ?Term1, ?Term2)
%
%  Unifies arguments 1..Num of Term1 and Term2.

same_args(0, _, _) :-
	!.

same_args(N, F1, F2) :-
	arg(N, F1, Arg),
	arg(N, F2, Arg),
	N1 is N - 1,
	same_args(N1, F1, F2).



fix_wrapped_code([], _) :- 
	!.

fix_wrapped_code([A|B], [call_tac(_, _, A)|WB]) :-
	fix_wrapped_code(B, WB).


/*
** 
*/

call_tac(N, L, C) :- 
	write('******  '), write_wrapper(L), nl,
	wrap(N, L, call_tac(C)).

/*
**
*/

call_tac(A) :-
	nonvar(A),
	!,
	(
	    ctac(A)
	;
	    report_failure(A)
	),
	!.

	
/*
call_tac(A) :-	
	is_focus(F),
	is_goal(G),
	F == G,
	complete_retry_delays, 
	(var(A) -> writeln('Focus = Goal, Quit')),	
	A = not_used, !.
*/

call_tac(X) :-	
	try_heuristics(Action),
	!,
	(
	    (Action == quit -> X = not_used)
	;
	    call_tac(X)
	).

call_tac(X) :- 
	trial_reportIt,
	!,
	it(X).

/*
**  The de-abbreviating, backtracking interactive tactic with 
**  multiple heuristic levels
*/

it(X) :- 
        ipLookup(cw, CW),
        ((ipLookup(in_hyp, true) ->
	    Prompt = [nl,CW,'h% '])
	;
	    (Prompt = [nl,CW,'% '])
	),
	get_command(Prompt, C),
	C \== undo,
	(
	    (get_hint(C), 
	     de_abbv(C, Command),
	     it(Command, X))
	; 
	    (trial_reportIt,
	     it(X))
	),
	!.




it(do_suspend,  fail).
it(quit,  not_used) :-
	!, 
	check_or_writelist(
		is_justified_focus,
	    	['Use really_quit to exit unjustified window\n']).

it(really_quit, not_used).
it(Line, 	(Line, Z)) :-
	wrap_tac(Line),
	!,
	call_tac(Z).

ct(do_suspend,  fail, true, true).
ct(Line, 	(Line, Z), Line, Z).



get_hint(Line) :- nonvar(Line), !.
get_hint(Line) :- it(Line).

/*
** Low level replacement for call/1
*/

ctac(X) :- var(X), !, call_tac(X).
ctac((A, B)) :- !, ctac(A), ctac(B).
ctac(X) :- call(X).

/*
**
*/

report_failure(X) :-
        quiet_mode, !,
        writeln('     : failed'),
        fail.

report_failure(X) :-
        !, nl,
        writeR(X, _), nl,
        writeln('     : failed'),
        nl,
        fail.

